# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from typing import Any

from flask import Blueprint, current_app, session, url_for

from bl_hector.infrastructure.flask import services
from bl_hector.infrastructure.flask.utils import presenter_to_response
from bl_hector.interfaces.to_http import Redirection, as_html

blueprint = Blueprint("auth", __name__)


@blueprint.get("")
@presenter_to_response
def root() -> Any:
    return Redirection(url_for("books.search"))


@blueprint.get("/login")
@presenter_to_response
def login() -> Any:
    auth_links = current_app.auth_links  # type: ignore[attr-defined]
    if not auth_links:
        return Redirection(url_for("aliases.root"))
    if len(auth_links) == 1:
        return Redirection(url_for(auth_links[0]["route"]))
    return as_html.SimplePresenter(
        "auth/login", links=auth_links, user=services.get_user()
    )


@blueprint.get("/logout")
@presenter_to_response
def logout() -> Any:
    session.clear()
    return Redirection(url_for("aliases.root"))



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from typing import Any, cast
from uuid import uuid4

import pyotp
from flask import Blueprint, flash, request, session, url_for

from bl_hector.domain.administration.entities import User
from bl_hector.domain.administration.value_objects import UserId
from bl_hector.infrastructure import ONLY_USER
from bl_hector.infrastructure.flask import services
from bl_hector.infrastructure.flask.utils import presenter_to_response
from bl_hector.infrastructure.settings import TotpSettings
from bl_hector.interfaces.to_http import Redirection, as_html

blueprint = Blueprint("totp", __name__)


@blueprint.get("/login")
@presenter_to_response
def login() -> Any:
    return as_html.SimplePresenter("totp/login", user=services.get_user())


@blueprint.post("/login")
@presenter_to_response
def login_POST() -> Any:
    # TOTP settings must be defined for this route to be accessible.
    # Other alternatives: 1) check presence with `if` or 2) ignore type.
    TOTP = cast(TotpSettings, services.get_settings().TOTP)

    totp = pyotp.TOTP(TOTP.SECRET)
    if not totp.verify(request.form.get("password", "")):
        flash("Error authenticating with TOTP.", "error")
        return Redirection(url_for("totp.login"))

    if not (user := services.get_users().by_login(ONLY_USER)):
        user = User(UserId(str(uuid4())), ONLY_USER, "The only user")
        services.get_users().add(user)

    session["user_id"] = user.id

    flash("Success authenticating with TOTP.", "success")
    return Redirection(url_for("aliases.root"))


@blueprint.get("/logout")
@presenter_to_response
def logout() -> Any:
    return Redirection(url_for("auth.logout"))



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from typing import Any

from flask import Blueprint, flash, get_flashed_messages, request

from bl_hector.application.use_cases import (
    add_book,
    display_book,
    look_up_book,
    search_books,
    update_book,
)
from bl_hector.infrastructure.flask import services
from bl_hector.infrastructure.flask.utils import htmx, presenter_to_response
from bl_hector.interfaces import from_dict as controllers
from bl_hector.interfaces.to_http import as_html as presenters

blueprint = Blueprint("books", __name__)


def notify(message: str, type: str) -> None:
    get_flashed_messages()  # FIXME… wtf?!
    flash(message, type)


@blueprint.get("")
@presenter_to_response
def search() -> Any:
    presenter: search_books.Presenter
    if htmx:
        presenter = presenters.SearchBooksFragment(
            htmx.target, request.args, user=services.get_user()
        )
    else:
        presenter = presenters.SearchBooks(request.args, user=services.get_user())
    interactor = search_books.Interactor(presenter, services.get_books())
    controller = controllers.SearchBooks(request.args)
    controller.call(interactor)
    return presenter


@blueprint.post("__validate__")
@presenter_to_response
def validate() -> Any:
    return presenters.ValidateBook(request.form, user=services.get_user())


@blueprint.get("__add__")
@presenter_to_response
def add_form() -> Any:
    return presenters.AddBookForm(user=services.get_user())


@blueprint.post("__add__")
@presenter_to_response
def add() -> Any:
    presenter: add_book.Presenter
    if htmx:
        presenter = presenters.AddBookFragment(
            htmx.target, request.form, notify, user=services.get_user()
        )
    else:
        presenter = presenters.AddBook(request.form, notify, user=services.get_user())
    interactor = add_book.Interactor(
        presenter,
        services.get_books(),
        services.get_calendar(),
        services.get_permissions(),
    )
    controller = controllers.AddBook(request.form, str(services.get_user().id))
    controller.call(interactor)
    return presenter


@blueprint.post("__look_up__")
@presenter_to_response
def look_up() -> Any:
    presenter = presenters.LookUpBook(user=services.get_user())
    interactor = look_up_book.Interactor(
        presenter, services.get_book_info_provider(), services.get_cover_provider()
    )
    interactor.execute(look_up_book.Request(request.form.get("isbn", "")))
    return presenter


@blueprint.get("<string:isbn>")
@presenter_to_response
def display(isbn: str) -> Any:
    presenter = presenters.DisplayBook(notify, user=services.get_user())
    interactor = display_book.Interactor(presenter, services.get_books())
    interactor.execute(display_book.Request(isbn))
    return presenter


@blueprint.get("<string:isbn>/__update__")
@presenter_to_response
def update_form(isbn: str) -> Any:
    presenter = presenters.UpdateBookForm(isbn, notify, user=services.get_user())
    interactor = display_book.Interactor(presenter, services.get_books())
    interactor.execute(display_book.Request(isbn))
    return presenter


@blueprint.post("<string:isbn>/__update__")
@presenter_to_response
def update(isbn: str) -> Any:
    presenter: update_book.Presenter
    if htmx:
        presenter = presenters.UpdateBookFragment(
            htmx.target, isbn, request.form, notify, user=services.get_user()
        )
    else:
        presenter = presenters.UpdateBook(
            isbn, request.form, notify, user=services.get_user()
        )
    interactor = update_book.Interactor(
        presenter,
        services.get_books(),
        services.get_calendar(),
        services.get_permissions(),
    )
    controller = controllers.UpdateBook(isbn, request.form, str(services.get_user().id))
    controller.call(interactor)
    return presenter



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import sys

import typer
from typing_extensions import Annotated

from bl_hector.application.use_cases import add_book, display_book, look_up_book
from bl_hector.infrastructure import ONLY_USER, DummyPermissions, SystemCalendar
from bl_hector.infrastructure.typer import services
from bl_hector.interfaces import from_json as controllers
from bl_hector.interfaces.to_terminal import LookUpBookInterface, as_json, as_text

cmd = typer.Typer()


@cmd.command()
def display(ctx: typer.Context, isbn: str) -> None:
    """Display a book from the collection."""
    with services.get_books(ctx.obj) as books:
        presenter = as_text.DisplayBook(
            lambda m: typer.echo(m), translator=services.get_translator()
        )
        interactor = display_book.Interactor(presenter, books)
        interactor.execute(display_book.Request(isbn))
        raise typer.Exit(presenter.exit_code())


@cmd.command()
def look_up(
    ctx: typer.Context,
    isbn: str,
    json: Annotated[bool, typer.Option(help="Format output in JSON.")] = False,
) -> None:
    """Look up for book information based on its ISBN number."""
    presenter: LookUpBookInterface
    if json:
        presenter = as_json.LookUpBook(lambda m: typer.echo(m))
    else:
        presenter = as_text.LookUpBook(
            lambda m: typer.echo(m), translator=services.get_translator()
        )
    interactor = look_up_book.Interactor(
        presenter, services.get_book_info_provider(), services.get_cover_provider()
    )
    interactor.execute(look_up_book.Request(isbn))
    raise typer.Exit(presenter.exit_code())


@cmd.command()
def add(ctx: typer.Context) -> None:
    """Add a book to the collection."""
    calendar = SystemCalendar()
    permissions = DummyPermissions()
    with services.get_books(ctx.obj) as books:
        presenter = as_text.AddBook(
            lambda m: typer.echo(m), translator=services.get_translator()
        )
        interactor = add_book.Interactor(presenter, books, calendar, permissions)
        controller = controllers.AddBook(sys.stdin.readline(), ONLY_USER)
        controller.call(interactor)
        raise typer.Exit(presenter.exit_code())



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json
from http import HTTPStatus as HTTP
from typing import Any

from bl_hector.interfaces.to_http import HttpPresenter


class Dict(HttpPresenter):
    def __init__(self, data: dict[str, Any], /, *, status_code: HTTP = HTTP.OK) -> None:
        self.__status_code = status_code
        self.__headers = {"Content-Type": "application/json"}
        self.__data = data

    def status_code(self) -> int:
        return int(self.__status_code)

    def headers(self) -> dict[str, str]:
        return self.__headers

    def data(self) -> str:
        return json.dumps(self.__data)



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json
import logging

from bl_hector.application.use_cases import add_book


class AddBook:
    def __init__(self, data: str, user_id: str = "") -> None:
        self.__data = data
        self.__user_id = user_id

    def call(self, interactor: add_book.Interactor) -> None:
        if not self.__data:
            return

        try:
            data = json.loads(self.__data)
        except Exception as exc:
            logging.error(f"AddBook: {exc}")
            logging.exception(exc)
            data = {}

        interactor.execute(
            add_book.Request(
                user_id=self.__user_id,
                isbn=str(data.get("isbn", "")),
                title=str(data.get("title", "")),
                year=int(data.get("year", "0")),
                authors=data.get("authors", []),
                genres=data.get("genres", []),
                cover=str(data.get("cover", "")),
            )
        )



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import logging
from typing import Optional

import webauthn
from webauthn.helpers.structs import (
    AuthenticationCredential,
    PublicKeyCredentialDescriptor,
    RegistrationCredential,
)

from bl_hector.domain.administration.entities import (
    Challenge,
    Credential,
    RelyingParty,
    User,
)


def credential_creation_options(
    rp: RelyingParty, user: User, challenge: Challenge
) -> str:
    pcco = webauthn.generate_registration_options(
        rp_id=rp.id,
        rp_name=rp.name,
        user_id=str(user.id),
        user_name=user.name,
        challenge=bytes(challenge),
    )
    return webauthn.options_to_json(pcco)


def parse_credentials(
    data: bytes, challenge: Challenge, rp: RelyingParty, user: User
) -> Optional[Credential]:
    try:
        verification = webauthn.verify_registration_response(
            credential=RegistrationCredential.parse_raw(data),
            expected_challenge=bytes(challenge),
            expected_origin=rp.origin,
            expected_rp_id=rp.id,
        )
        return Credential(
            verification.credential_id, verification.credential_public_key, user.id
        )
    except Exception as exc:
        logging.exception(exc)
        return None


def authentication_options(
    rp: RelyingParty, user: User, credentials: list[Credential], challenge: Challenge
) -> str:
    allowed_credentials = [
        PublicKeyCredentialDescriptor(id=credential.id) for credential in credentials
    ]
    authentication_options = webauthn.generate_authentication_options(
        rp_id=rp.id,
        allow_credentials=allowed_credentials,
        challenge=bytes(challenge),
    )
    return webauthn.options_to_json(authentication_options)


def verify_credentials(
    data: bytes,
    challenge: Challenge,
    rp: RelyingParty,
    user: User,
    credentials: list[Credential],
) -> bool:
    try:
        auth_credential = AuthenticationCredential.parse_raw(data)
    except Exception as exc:
        logging.exception(exc)
        return False

    auth_id = webauthn.base64url_to_bytes(auth_credential.id)
    if not (credential := next((c for c in credentials if c.id == auth_id), None)):
        logging.info("No matching credentials.")
        return False

    webauthn.verify_authentication_response(
        credential=auth_credential,
        expected_challenge=bytes(challenge),
        expected_origin=rp.origin,
        expected_rp_id=rp.id,
        credential_public_key=credential.public_key,
        credential_current_sign_count=0,
    )
    return True



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import datetime
import secrets
from uuid import uuid4

from .entities import Challenge, User
from .value_objects import ChallengeId, ChallengeValue

SECOND = 1
MINUTE = 60 * SECOND
DEFAULT_LIFE_TIME = 10 * MINUTE


def create_challenge_for(user: User) -> Challenge:
    return Challenge(
        ChallengeId(str(uuid4())),
        ChallengeValue(secrets.token_hex(16)),
        datetime.datetime.now() + datetime.timedelta(minutes=DEFAULT_LIFE_TIME),
        user.id,
    )



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from pypugjs.ext.jinja import Compiler, PyPugJSExtension
from pypugjs.parser import Parser
from pypugjs.utils import process


class PatchedCompiler(Compiler):  # type: ignore
    def visitInclude(self, node):  # type: ignore
        src, _, _ = self.options["loader"].get_source(
            self.options["environment"], self.format_path(node.path)
        )
        parser = Parser(src)
        block = parser.parse()
        return self.visit(block)


class PatchedPyPugJSExtension(PyPugJSExtension):  # type: ignore
    def preprocess(self, source, name, filename=None):  # type: ignore
        options = {
            "environment": self.environment,
            "loader": self.environment.loader,
        }

        return process(source, filename=name, compiler=PatchedCompiler, **options)



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from dataclasses import dataclass, fields
from typing import Optional

from bl_hector.domain.collection_management.entities import Book
from bl_hector.domain.collection_management.errors import InvalidValue
from bl_hector.domain.collection_management.repositories import Books
from bl_hector.domain.collection_management.value_objects import (
    Author,
    Genre,
    Isbn,
    Title,
    Year,
)


@dataclass(frozen=True)
class Request:
    isbn: Optional[str] = None
    title: Optional[str] = None
    year: Optional[int] = None
    author: Optional[str] = None
    genre: Optional[str] = None
    page_number: Optional[int] = None
    page_size: Optional[int] = None


@dataclass
class Errors:
    isbn: Optional[InvalidValue] = None
    title: Optional[InvalidValue] = None
    year: Optional[InvalidValue] = None
    author: Optional[InvalidValue] = None
    genre: Optional[InvalidValue] = None

    def __bool__(self) -> bool:
        return any([getattr(self, f.name) for f in fields(self)])


class Presenter(ABC):
    @abstractmethod
    def bad_request(self, errors: Errors) -> None:
        ...

    @abstractmethod
    def book(self, book: Book) -> None:
        ...


@dataclass(frozen=True)
class Interactor:
    presenter: Presenter
    books: Books

    def execute(self, request: Request) -> None:
        if errors := self.__validate_request(request):
            return self.presenter.bad_request(errors)

        books = self.books.search(
            isbn=Isbn.instanciate(request.isbn) if request.isbn else None,
            title=Title.instanciate(request.title) if request.title else None,
            year=Year.instanciate(request.year) if request.year else None,
            author=Author.instanciate(request.author) if request.author else None,
            genre=Genre.instanciate(request.genre) if request.genre else None,
            page_number=request.page_number,
            page_size=request.page_size,
        )
        for book in books:
            self.presenter.book(book)

    def __validate_request(self, request: Request) -> Errors:
        errors = Errors()
        if request.isbn:
            try:
                Isbn.instanciate(request.isbn)
            except InvalidValue as exc:
                errors.isbn = exc
        if request.title:
            try:
                Title.instanciate(request.title)
            except InvalidValue as exc:
                errors.title = exc
        if request.year is not None:
            try:
                Year.instanciate(request.year)
            except InvalidValue as exc:
                errors.year = exc
        if request.author:
            try:
                Author.instanciate(request.author)
            except InvalidValue as exc:
                errors.author = exc
        if request.genre:
            try:
                Genre.instanciate(request.genre)
            except InvalidValue as exc:
                errors.genre = exc
        return errors



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import datetime
from abc import ABC, abstractmethod
from typing import Optional

from .entities import Challenge, Credential, User
from .enumerations import Permissions as P
from .value_objects import UserId


class Users(ABC):
    @abstractmethod
    def by_login(self, login: str) -> Optional[User]:
        ...

    @abstractmethod
    def add(self, user: User) -> None:
        ...


class Credentials(ABC):
    @abstractmethod
    def for_user(self, user: User) -> list[Credential]:
        ...

    @abstractmethod
    def add(self, credential: Credential) -> None:
        ...


class Challenges(ABC):
    @abstractmethod
    def valid_for(self, user: User, now: datetime.datetime) -> Optional[Challenge]:
        ...

    @abstractmethod
    def add(self, challenge: Challenge) -> None:
        ...


class Permissions(ABC):
    @abstractmethod
    def is_authorized_to(self, user_id: UserId, permission: P) -> bool:
        ...



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import datetime
from typing import Any, Optional

from sqlalchemy import Column
from sqlalchemy import Date as DateType
from sqlalchemy import (
    DateTime,
    LargeBinary,
    MetaData,
    SmallInteger,
    String,
    Table,
    delete,
    insert,
    select,
    update,
)
from sqlalchemy.engine import Connection

from bl_hector.domain.administration.entities import Challenge, Credential, User
from bl_hector.domain.administration.repositories import Challenges as IChallenges
from bl_hector.domain.administration.repositories import Credentials as ICredentials
from bl_hector.domain.administration.repositories import Users as IUsers
from bl_hector.domain.administration.value_objects import (
    ChallengeId,
    ChallengeValue,
    UserId,
)
from bl_hector.domain.collection_management.entities import Book
from bl_hector.domain.collection_management.errors import UnknownBook
from bl_hector.domain.collection_management.repositories import Books as IBooks
from bl_hector.domain.collection_management.value_objects import (
    Author,
    Cover,
    Date,
    Genre,
    Isbn,
    Title,
    Year,
)

META_DATA = MetaData()
DEFAULT_PAGE_SIZE = 100


books = Table(
    "books",
    META_DATA,
    Column("isbn", String(Isbn.MAX), primary_key=True),
    Column("title", String(Title.MAX)),
    Column("year", SmallInteger()),
    Column("authors", String()),
    Column("genres", String()),
    Column("cover", String()),
    Column("added_on", DateType()),
    Column("updated_on", DateType()),
)


class Books(IBooks):
    def __init__(self, connection: Connection) -> None:
        self.__connection = connection

    def by_isbn(self, isbn: Isbn) -> Optional[Book]:
        stmt = select(books).where(books.c.isbn == str(isbn)).limit(1)
        if row := self.__connection.execute(stmt).first():
            return self.__row_to_book(row)
        return None

    def search(
        self,
        /,
        *,
        isbn: Optional[Isbn] = None,
        title: Optional[Title] = None,
        year: Optional[Year] = None,
        author: Optional[Author] = None,
        genre: Optional[Genre] = None,
        page_number: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> list[Book]:
        stmt = select(books)
        if isbn:
            stmt = stmt.where(books.c.isbn == str(isbn))
        if title:
            stmt = stmt.where(books.c.title.contains(str(title)))
        if year:
            stmt = stmt.where(books.c.year == int(year))
        if author:
            stmt = stmt.where(books.c.authors.contains(str(author)))
        if genre:
            stmt = stmt.where(books.c.genres.contains(str(genre)))

        stmt = stmt.order_by(books.c.year.desc()).order_by(books.c.title)

        if page_number:
            size = page_size or DEFAULT_PAGE_SIZE
            stmt = stmt.offset((page_number - 1) * size)
            stmt = stmt.limit(size)

        return [self.__row_to_book(row) for row in self.__connection.execute(stmt)]

    def __row_to_book(self, row: Any) -> Book:
        return Book(
            Date(row.added_on),
            Date(row.updated_on),
            Isbn(row.isbn),
            Title(row.title),
            Year(row.year),
            [Author(r.strip()) for r in row.authors.split(",") if r],
            [Genre(r.strip()) for r in row.genres.split(",") if r],
            Cover(row.cover) if row.cover else None,
        )

    def add(self, book: Book) -> None:
        stmt = insert(books).values(
            added_on=book.added_on.to_date(),
            updated_on=book.updated_on.to_date(),
            isbn=str(book.isbn),
            title=str(book.title),
            year=int(book.year),
            authors=", ".join([str(a) for a in book.authors]),
            genres=", ".join([str(g) for g in book.genres]),
            cover=str(book.cover) if book.cover else None,
        )
        self.__connection.execute(stmt)

    def update(self, book: Book) -> None:
        if not self.by_isbn(book.isbn):
            raise UnknownBook(str(book.isbn))
        stmt = (
            update(books)
            .values(
                updated_on=book.updated_on.to_date(),
                title=str(book.title),
                year=int(book.year),
                authors=", ".join([str(a) for a in book.authors]),
                genres=", ".join([str(g) for g in book.genres]),
                cover=str(book.cover) if book.cover else None,
            )
            .where(books.c.isbn == str(book.isbn))
        )
        self.__connection.execute(stmt)


users = Table(
    "users",
    META_DATA,
    Column("id", String(UserId.MAX), primary_key=True),
    Column("login", String()),
    Column("name", String()),
)


class Users(IUsers):
    def __init__(self, connection: Connection) -> None:
        self.__connection = connection

    def by_login(self, login: str) -> Optional[User]:
        stmt = select(users).where(users.c.login == login).limit(1)
        if row := self.__connection.execute(stmt).first():
            return User(UserId(row.id), row.login, row.name)
        return None

    def add(self, user: User) -> None:
        stmt = insert(users).values(
            id=str(user.id),
            login=user.login,
            name=user.name,
        )
        self.__connection.execute(stmt)


credentials = Table(
    "credentials",
    META_DATA,
    Column("id", LargeBinary(), primary_key=True),
    Column("public_key", LargeBinary()),
    Column("user_id", String(UserId.MAX)),
)


class Credentials(ICredentials):
    def __init__(self, connection: Connection) -> None:
        self.__connection = connection

    def for_user(self, user: User) -> list[Credential]:
        stmt = select(credentials).where(credentials.c.user_id == str(user.id))
        return [
            Credential(row.id, row.public_key, UserId(row.user_id))
            for row in self.__connection.execute(stmt).all()
        ]

    def add(self, creds: Credential) -> None:
        stmt = insert(credentials).values(
            id=creds.id,
            public_key=creds.public_key,
            user_id=str(creds.user_id),
        )
        self.__connection.execute(stmt)


challenges = Table(
    "challenges",
    META_DATA,
    Column("id", String(ChallengeId.MAX), primary_key=True),
    Column("value", String(ChallengeValue.MAX)),
    Column("expiration", DateTime()),
    Column("user_id", String(UserId.MAX)),
)


class Challenges(IChallenges):
    def __init__(self, connection: Connection) -> None:
        self.__connection = connection

    def add(self, challenge: Challenge) -> None:
        self.__clean_up_for(challenge.user_id)

        stmt = insert(challenges).values(
            id=str(challenge.id),
            value=str(challenge.value),
            expiration=challenge.expiration,
            user_id=str(challenge.user_id),
        )
        self.__connection.execute(stmt)

    def __clean_up_for(self, user_id: UserId) -> None:
        stmt = delete(challenges).where(challenges.c.user_id == str(user_id))
        self.__connection.execute(stmt)

    def valid_for(self, user: User, now: datetime.datetime) -> Optional[Challenge]:
        stmt = (
            select(challenges)
            .where(challenges.c.user_id == str(user.id))
            .where(challenges.c.expiration > now)
        )
        if result := self.__connection.execute(stmt).first():
            return Challenge(
                ChallengeId(result.id),
                ChallengeValue(result.value),
                result.expiration,
                UserId(result.user_id),
            )
        return None



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime as dt
from http import HTTPStatus as HTTP
from typing import Any
from uuid import uuid4

from flask import Blueprint, request, session, url_for

from bl_hector.domain.administration.entities import User
from bl_hector.domain.administration.services import create_challenge_for
from bl_hector.domain.administration.value_objects import UserId
from bl_hector.infrastructure import ONLY_USER
from bl_hector.infrastructure.flask import services
from bl_hector.infrastructure.flask.utils import presenter_to_response
from bl_hector.interfaces.to_http import Redirection, as_html, as_json

from . import security

blueprint = Blueprint("webauthn", __name__)


@blueprint.get("/login")
@presenter_to_response
def login() -> Any:
    # For the time being, Hector only has one user with one credential
    if not (user := services.get_users().by_login(ONLY_USER)):
        return Redirection(url_for("webauthn.register"))

    if not (credentials := services.get_credentials().for_user(user)):
        return Redirection(url_for("webauthn.register"))

    challenge = create_challenge_for(user)
    services.get_challenges().add(challenge)

    return as_html.SimplePresenter(
        "webauthn/login",
        user=services.get_user(),
        options=security.authentication_options(
            services.get_relying_party(), user, credentials, challenge
        ),
    )


@blueprint.post("/verify-credential")
@presenter_to_response
def verify_credential() -> Any:
    if not (user := services.get_users().by_login(ONLY_USER)):
        return as_json.Dict({"verified": False}, status_code=HTTP.BAD_REQUEST)

    if not (challenge := services.get_challenges().valid_for(user, dt.now())):
        return as_json.Dict({"verified": False}, status_code=HTTP.BAD_REQUEST)

    if not (credentials := services.get_credentials().for_user(user)):
        return as_json.Dict({"verified": False}, status_code=HTTP.BAD_REQUEST)

    if not security.verify_credentials(
        request.get_data(),
        challenge,
        services.get_relying_party(),
        user,
        credentials,
    ):
        return as_json.Dict({"verified": False}, status_code=HTTP.BAD_REQUEST)

    session["user_id"] = user.id
    return as_json.Dict({"verified": True}, status_code=HTTP.OK)


@blueprint.get("/register")
@presenter_to_response
def register() -> Any:
    if not (user := services.get_users().by_login(ONLY_USER)):
        user = User(UserId(str(uuid4())), ONLY_USER, "The only user")
        services.get_users().add(user)

    if services.get_credentials().for_user(user):
        return Redirection(url_for("webauthn.login"))

    challenge = create_challenge_for(user)
    services.get_challenges().add(challenge)

    return as_html.SimplePresenter(
        "webauthn/register",
        user=services.get_user(),
        options=security.credential_creation_options(
            services.get_relying_party(), user, challenge
        ),
    )


@blueprint.post("/add-credential")
@presenter_to_response
def add_credential() -> Any:
    if not (user := services.get_users().by_login(ONLY_USER)):
        return as_json.Dict({"verified": False}, status_code=HTTP.BAD_REQUEST)

    if not (challenge := services.get_challenges().valid_for(user, dt.now())):
        return as_json.Dict({"verified": False}, status_code=HTTP.BAD_REQUEST)

    if not (
        credential := security.parse_credentials(
            request.get_data(), challenge, services.get_relying_party(), user
        )
    ):
        return as_json.Dict({"verified": False}, status_code=HTTP.BAD_REQUEST)

    services.get_credentials().add(credential)
    return as_json.Dict({"verified": True}, status_code=HTTP.CREATED)


@blueprint.get("/logout")
@presenter_to_response
def logout() -> Any:
    return Redirection(url_for("auth.logout"))



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from typing import Callable, Optional

from bl_hector.application.use_cases import add_book, display_book
from bl_hector.domain.collection_management.entities import Book
from bl_hector.domain.collection_management.errors import InvalidValue
from bl_hector.domain.collection_management.value_objects import Isbn
from bl_hector.interfaces import translate_error
from bl_hector.interfaces.l10n import DummyTranslator, Translator
from bl_hector.interfaces.to_terminal import ExitCodes, LookUpBookInterface


class DisplayBook(display_book.Presenter):
    def __init__(
        self,
        printer: Callable[[str], None],
        /,
        *,
        translator: Translator = DummyTranslator(),
    ) -> None:
        self.__printer = printer
        self.__exit_code = ExitCodes.USAGE
        self._ = translator
        self.__exit_code = ExitCodes.NOT_FOUND

    def not_an_isbn(self, isbn: str) -> None:
        self.__exit_code = ExitCodes.BAD_REQUEST
        self.__printer(self._("not-an-isbn"))

    def book_not_found(self, isbn: Isbn) -> None:
        self.__exit_code = ExitCodes.NOT_FOUND
        self.__printer(self._("book-not-found"))

    def see_other(self, isbn: Isbn) -> None:
        pass

    def book(self, book: Book) -> None:
        self.__exit_code = ExitCodes.OK
        self.__info(self._("book-isbn"), str(book.isbn))
        self.__info(self._("book-title"), str(book.title))
        self.__info(self._("book-year"), str(book.year))
        self.__info(self._("book-authors"), ", ".join([str(b) for b in book.authors]))
        if book.genres:
            self.__info(self._("book-genres"), ", ".join([str(g) for g in book.genres]))

    def __info(self, name: str, value: str) -> None:
        self.__printer(self._("info-line", name=name, value=value))

    def exit_code(self) -> int:
        return self.__exit_code.value


class LookUpBook(LookUpBookInterface):
    def __init__(
        self,
        printer: Callable[[str], None],
        /,
        *,
        translator: Translator = DummyTranslator(),
    ) -> None:
        self.__printer = printer
        self.__exit_code = ExitCodes.USAGE
        self._ = translator

    def not_an_isbn(self, isbn: str) -> None:
        self.__exit_code = ExitCodes.BAD_REQUEST
        self.__printer(self._("not-an-isbn"))

    def book_not_found(self, isbn: Isbn) -> None:
        self.__exit_code = ExitCodes.NOT_FOUND
        self.__printer(self._("book-not-found"))

    def book(self, book: Book) -> None:
        self.__exit_code = ExitCodes.OK
        self.__info(self._("book-isbn"), str(book.isbn))
        self.__info(self._("book-title"), str(book.title))
        self.__info(self._("book-year"), str(book.year))
        self.__info(self._("book-authors"), ", ".join([str(b) for b in book.authors]))
        if book.genres:
            self.__info(self._("book-genres"), ", ".join([str(g) for g in book.genres]))

    def __info(self, name: str, value: str) -> None:
        self.__printer(self._("info-line", name=name, value=value))

    def exit_code(self) -> int:
        return self.__exit_code.value


class AddBook(add_book.Presenter):
    def __init__(
        self,
        printer: Callable[[str], None],
        /,
        *,
        translator: Translator = DummyTranslator(),
    ) -> None:
        self.__printer = printer
        self.__exit_code = ExitCodes.USAGE
        self._ = translator

    def not_authorized(self) -> None:
        self.__exit_code = ExitCodes.USAGE
        self.__printer(self._("access-not-authorized"))

    def bad_request(self, errors: add_book.Errors) -> None:
        self.__exit_code = ExitCodes.BAD_REQUEST
        self.__printer(self._("book-cannot-be-added"))
        self.__print_error("book-isbn", errors.isbn)
        self.__print_error("book-title", errors.title)
        self.__print_error("book-year", errors.year)
        self.__print_error("book-authors", errors.authors)

    def __print_error(self, name: str, error: Optional[InvalidValue]) -> None:
        if not error:
            return
        self.__printer(
            self._("info-line", name=self._(name), value=translate_error(self._, error))
        )

    def book_already_exists(self, book: Book) -> None:
        self.__exit_code = ExitCodes.BAD_REQUEST
        self.__printer(self._("book-already-exists"))

    def book_added(self, book: Book) -> None:
        self.__exit_code = ExitCodes.OK
        self.__printer(self._("book-added-text", isbn=str(book.isbn)))

    def exit_code(self) -> int:
        return self.__exit_code.value



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import logging
from contextlib import contextmanager
from typing import Any, Iterator

from sqlalchemy import create_engine
from typer import Exit

from bl_hector.infrastructure.isbnlib import BookInfoProvider
from bl_hector.infrastructure.requests import CoverProvider
from bl_hector.infrastructure.settings import CliSettings
from bl_hector.infrastructure.sqlalchemy.repositories import Books
from bl_hector.interfaces import l10n


def get_book_info_provider() -> BookInfoProvider:
    return BookInfoProvider()


def get_cover_provider() -> CoverProvider:
    return CoverProvider()


@contextmanager
def get_books(settings: CliSettings) -> Iterator[Books]:
    options: dict[str, Any] = {}
    if settings.DEBUG_SQL:
        options["echo"] = True
        options["echo_pool"] = "debug"

    engine = create_engine(settings.DSN, **options)
    connection = engine.connect()

    try:
        yield Books(connection)
    except Exit:
        # Raising an `Exit` is the way to end a typer command.
        connection.commit()
        raise
    except Exception as exc:
        logging.exception(exc)
        connection.rollback()
    else:
        connection.commit()


def get_translator() -> l10n.Translator:
    def wrapper(message_id: str, **kwargs: Any) -> str:
        return l10n.localization(l10n.DEFAULT_LOCALE).format_value(message_id, kwargs)

    return wrapper



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bl_hector',
 'bl_hector.application',
 'bl_hector.application.use_cases',
 'bl_hector.configuration',
 'bl_hector.domain',
 'bl_hector.domain.administration',
 'bl_hector.domain.collection_management',
 'bl_hector.infrastructure',
 'bl_hector.infrastructure.flask',
 'bl_hector.infrastructure.flask.aliases',
 'bl_hector.infrastructure.flask.auth',
 'bl_hector.infrastructure.flask.books',
 'bl_hector.infrastructure.flask.totp',
 'bl_hector.infrastructure.flask.webauthn',
 'bl_hector.infrastructure.isbnlib',
 'bl_hector.infrastructure.requests',
 'bl_hector.infrastructure.sqlalchemy',
 'bl_hector.infrastructure.typer',
 'bl_hector.interfaces',
 'bl_hector.interfaces.l10n',
 'bl_hector.interfaces.to_http',
 'bl_hector.interfaces.to_http.as_html',
 'bl_hector.interfaces.to_http.as_json',
 'bl_hector.interfaces.to_terminal']

package_data = \
{'': ['*'],
 'bl_hector.infrastructure.flask': ['static/*',
                                    'static/css/*',
                                    'static/img/placeholders/*',
                                    'static/js/*',
                                    'static/webfonts/*'],
 'bl_hector.interfaces.l10n': ['en-GB/*', 'fr-FR/*'],
 'bl_hector.interfaces.to_http.as_html': ['templates/*',
                                          'templates/auth/*',
                                          'templates/books/*',
                                          'templates/mixins/*',
                                          'templates/totp/*',
                                          'templates/webauthn/*']}

install_requires = \
['Flask>=2.3.2,<3.0.0',
 'SQLAlchemy>=2.0.15,<3.0.0',
 'bl-seth>=0.2.0,<0.3.0',
 'bl3d>=0.4.0,<0.5.0',
 'fluent.runtime>=0.4.0,<0.5.0',
 'isbnlib>=3.10.14,<4.0.0',
 'jinja2-fragments>=0.3.0,<0.4.0',
 'pypugjs>=5.9.12,<6.0.0',
 'requests>=2.31.0,<3.0.0',
 'typer>=0.9.0,<0.10.0']

extras_require = \
{'totp': ['pyotp>=2.8.0,<3.0.0'], 'webauthn': ['webauthn>=1.8.1,<2.0.0']}

entry_points = \
{'console_scripts': ['hector = bl_hector.configuration.cli:cli']}

setup_kwargs = {
    'name': 'bl-hector',
    'version': '0.3.1',
    'description': 'A collection manager.',
    'long_description': '# Hector — a collection manager\n\n## Install\n\nHector is available on PyPI under the name `bl_hector`.\nTo install, just run `python -m pip install bl_hector`.\n\n\n## Configure\n\nHector is configured using environment variables.\nSee [the `settings` module](bl_hector/infrastructure/settings.py) for\na comprehensive list of configuration variables.\n\nAll the variable names must be prefixed with `HECTOR_`. For instance\xa0:\n\n```console\n# The secret can be generated using the `secrets.token_hex()` function.\n$ export HECTOR_SECRET_KEY="XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX"\n\n# Additional Python database drivers might be required depending on the DSN.\n$ export HECTOR_DSN="sqlite:///data.sqlite"\n```\n\n\n## Authentication\n\nTo enable WebAuthn authentication, you must install extra dependencies (`bl-hector[webauthn]`)\nand enable it explicitly:\n\n```console\n$ export HECTOR_WEBAUTHN_ENABLED=1\n```\n\nTOTP authentication is provided to be able to login on servers that do not (yet) support\nthe `cryptography` module. You must install extra dependencies (`bl-hector[totp]`)\nand enable it explicitly by setting a base32 random secret:\n\n```console\n# The secret can be generated using the `pyotp.random_base32()` function.\n$ export HECTOR_TOTP_SECRET=XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n```\n\nNote that it is a highly insecure way of authenticating, as anyone gaining access to your\nOTP generator would be able to login.\n\n\n## Initialise\n\nOnce configured, you must initialise Hector\'s database with the dedicated command:\n\n```console\n$ hector init-db\n```\n\n\n## Run\n\nHector being a Flask application, it can be run using any WSGI server,\nfor instance, with [Gunicorn](https://gunicorn.org):\n\n```console\n$ gunicorn --access-logfile="-" -w 4 -b 127.0.0.1:3000 "bl_hector.configuration.wsgi:app()"\n```\n\n\n## Contributing\n\nSee [CONTRIBUTING.md]() to set up a development environment.\n',
    'author': 'Tanguy Le Carrour',
    'author_email': 'tanguy@bioneland.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://git.easter-eggs.org/bioneland/hector',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from http import HTTPStatus as HTTP
from pathlib import Path
from typing import Any, Callable, Optional

from jinja2 import Environment, FileSystemLoader, select_autoescape
from jinja2_fragments import render_block
from werkzeug.datastructures import MultiDict

from bl_hector.application.use_cases import (
    add_book,
    display_book,
    look_up_book,
    search_books,
    update_book,
)
from bl_hector.domain.collection_management import validators
from bl_hector.domain.collection_management.entities import Book
from bl_hector.domain.collection_management.value_objects import Isbn
from bl_hector.interfaces import Pager, User, l10n, translate_error
from bl_hector.interfaces.to_http import HttpMeta, HttpPresenter

ENVIRONMENT = Environment(
    loader=FileSystemLoader([Path(__file__).parent / "templates"]),
    autoescape=select_autoescape(),
    extensions=["bl_hector.interfaces.utils.PatchedPyPugJSExtension"],
)


def register_jinja_global(key: str, value: Any) -> None:
    ENVIRONMENT.globals[key] = value


def url_for(*args: Any, **kwargs: Any) -> str:
    if "url_for" not in ENVIRONMENT.globals:
        raise RuntimeError("`url_for` is not declared on the environment!")
    return ENVIRONMENT.globals["url_for"](*args, **kwargs)  # type: ignore


class HtmlMeta(HttpMeta):
    def __init__(self, /, *, status: int = HTTP.OK, is_logged_in: bool = False) -> None:
        self.__status_code: int = status
        self.__headers: dict[str, Any] = {"Content-Type": "text/html; charset=UTF-8"}
        self.__is_logged_in = is_logged_in

    def see_other(self, target: str, /, *, permanent: bool = False) -> None:
        self.__status_code = HTTP.MOVED_PERMANENTLY if permanent else HTTP.SEE_OTHER

        if "Content-Type" in self.__headers:
            del self.__headers["Content-Type"]

        self.__headers["Location"] = target

    def bad_request(self) -> None:
        self.__status_code = HTTP.BAD_REQUEST

    def not_authorized(self) -> None:
        if self.__is_logged_in:
            self.__status_code = HTTP.FORBIDDEN
        else:
            self.__status_code = HTTP.UNAUTHORIZED

    def status_code(self) -> int:
        return self.__status_code

    def headers(self) -> dict[str, Any]:
        return self.__headers


class HtmxMeta(HttpMeta):
    def __init__(self, /, *, status: int = HTTP.OK, is_logged_in: bool = False) -> None:
        self.__status_code: int = status
        self.__headers: dict[str, Any] = {"Content-Type": "text/html; charset=UTF-8"}
        self.__is_logged_in = is_logged_in

    def see_other(self, target: str, /, *, permanent: bool = False) -> None:
        self.__status_code = HTTP.MOVED_PERMANENTLY if permanent else HTTP.SEE_OTHER

        if "Content-Type" in self.__headers:
            del self.__headers["Content-Type"]

        self.__headers["HX-Location"] = target

    def replace(self, target: str) -> None:
        self.__headers["HX-Push-Url"] = target

    def bad_request(self) -> None:
        self.__status_code = HTTP.BAD_REQUEST

    def not_authorized(self) -> None:
        if self.__is_logged_in:
            self.__status_code = HTTP.FORBIDDEN
        else:
            self.__status_code = HTTP.UNAUTHORIZED

    def status_code(self) -> int:
        return self.__status_code

    def headers(self) -> dict[str, Any]:
        # Prevent problem when navigating with the back button
        self.__headers["Cache-Control"] = "no-store, max-age=0"
        return self.__headers


class HtmlContent:
    def __init__(self, target: str, context: Optional[dict[str, Any]] = None) -> None:
        self.__template, self.__fragment = self.__parse(target)
        self.__context = context or {}

    def __parse(self, target: str) -> tuple[str, str]:
        e = target.split("#", 1)
        t, f = (e[0], "") if len(e) == 1 else (e[0], e[1])
        return t.removesuffix(".pug") + ".pug", f

    def __str__(self) -> str:
        if self.__fragment:
            return render_block(  # type: ignore
                ENVIRONMENT, self.__template, self.__fragment, **self.__context
            )
        return ENVIRONMENT.get_template(self.__template).render(**self.__context)

    def set(self, key: str, value: Any) -> None:
        self.__context[key] = value

    def append(self, key: str, value: Any) -> None:
        if key not in self.__context:
            self.__context[key] = []
        self.__context[key].append(value)

    def delete(self, key: str) -> None:
        if key in self.__context:
            del self.__context[key]

    def not_authorized(self) -> None:
        if "user" in self.__context:
            self.__context["error"] = self.__translate("access-forbidden")
        else:
            self.__context["error"] = self.__translate("access-not-authorized")

    def __translate(self, message: str) -> str:
        if "_" in self.__context:
            return str(self.__context["_"](message))
        return message


class HtmlOverHttpMixin:
    meta: HttpMeta
    content: HtmlContent

    def status_code(self) -> int:
        return self.meta.status_code()

    def headers(self) -> dict[str, str]:
        return self.meta.headers()

    def data(self) -> str:
        return str(self.content)

    def not_authorized(self) -> None:
        self.meta.not_authorized()
        self.content.not_authorized()


class SimplePresenter(HtmlOverHttpMixin, HttpPresenter):
    def __init__(
        self, template: str, /, *, user: Optional[User] = None, **context: Any
    ) -> None:
        self.meta = HtmlMeta()
        self.content = HtmlContent(
            template,
            {
                "user": user,
                "_": l10n.translator_for(user.locale if user else ""),
                **context,
            },
        )


class BadRequest(HtmlOverHttpMixin, HttpPresenter):
    def __init__(self, /, *, user: Optional[User] = None) -> None:
        _ = l10n.translator_for(user.locale if user else "")
        self.meta = HtmlMeta(status=HTTP.BAD_REQUEST)
        self.content = HtmlContent(
            "error", {"user": user, "_": _, "message": _("bad-request")}
        )


class SearchBooksOptionalFragment(
    HtmlOverHttpMixin, HttpPresenter, search_books.Presenter
):
    def __init__(
        self,
        data: MultiDict[str, Any],
        /,
        *,
        fragment: str = "",
        user: Optional[User] = None,
    ) -> None:
        self._ = l10n.translator_for(user.locale if user else "")
        pager = Pager(url_for("books.search", **data))

        if fragment:
            self.meta = HtmxMeta()
            self.meta.replace(
                url_for(
                    "books.search",
                    # Remove empty values and page number
                    **{k: v for k, v in data.items() if v and k != "page"},
                )
            )
        else:
            self.meta = HtmlMeta()

        self.content = HtmlContent(
            f"books/search#{fragment}",
            {"_": self._, "user": user, "data": data, "errors": {}, "pager": pager},
        )

        if data:  # Handle the case when the search returns no book.
            self.content.set("books", [])

    def bad_request(self, errors: search_books.Errors) -> None:
        self.content.set(
            "errors",
            {
                "isbn": translate_error(self._, errors.isbn),
                "title": translate_error(self._, errors.title),
                "year": translate_error(self._, errors.year),
                "author": translate_error(self._, errors.author),
                "genre": translate_error(self._, errors.genre),
            },
        )
        self.content.delete("books")

    def book(self, book: Book) -> None:
        self.content.append(
            "books",
            {
                "isbn": str(book.isbn),
                "title": str(book.title),
                "year": int(book.year),
                "authors": [str(a) for a in book.authors],
                "genres": [str(g) for g in book.genres],
                "cover": str(book.cover) if book.cover else "",
            },
        )


class SearchBooksFragment(SearchBooksOptionalFragment):
    def __init__(
        self,
        fragment: str,
        data: MultiDict[str, Any],
        /,
        *,
        user: Optional[User] = None,
    ) -> None:
        super().__init__(data, fragment=fragment, user=user)


class SearchBooks(SearchBooksOptionalFragment):
    def __init__(
        self, data: MultiDict[str, Any], /, *, user: Optional[User] = None
    ) -> None:
        super().__init__(data, user=user)


class AddBookForm(HtmlOverHttpMixin, HttpPresenter):
    def __init__(self, /, *, user: Optional[User] = None) -> None:
        self._ = l10n.translator_for(user.locale if user else "")
        self.meta = HtmlMeta(is_logged_in=bool(user))
        self.content = HtmlContent(
            "books/add", {"user": user, "_": self._, "data": {}, "errors": {}}
        )


class AddBookOptionalFragment(HtmlOverHttpMixin, HttpPresenter, add_book.Presenter):
    def __init__(
        self,
        data: MultiDict[str, Any],
        notify: Callable[[str, str], None],
        /,
        *,
        fragment: str = "",
        user: Optional[User] = None,
    ) -> None:
        self._ = l10n.translator_for(user.locale if user else "")
        if fragment:
            self.meta = HtmxMeta(is_logged_in=bool(user))
        else:
            self.meta = HtmlMeta(is_logged_in=bool(user))
        self.content = HtmlContent(
            "books/add",
            {"user": user, "_": self._, "data": data, "errors": {}},
        )
        self.__notify = notify

    def bad_request(self, errors: add_book.Errors) -> None:
        self.content.set(
            "errors",
            {
                "isbn": translate_error(self._, errors.isbn),
                "title": translate_error(self._, errors.title),
                "year": translate_error(self._, errors.year),
                "authors": translate_error(self._, errors.authors),
            },
        )

    def book_already_exists(self, book: Book) -> None:
        self.__notify(self._("book-already-exists"), "info")
        self.meta.see_other(url_for("books.display", isbn=str(book.isbn)))

    def book_added(self, book: Book) -> None:
        self.__notify(
            self._(
                "book-added-html",
                isbn=str(book.isbn),
                url=url_for("books.display", isbn=str(book.isbn)),
            ),
            "success",
        )
        self.meta.see_other(url_for("books.add"))


class AddBookFragment(AddBookOptionalFragment):
    def __init__(
        self,
        fragment: str,
        data: MultiDict[str, Any],
        notify: Callable[[str, str], None],
        /,
        *,
        user: Optional[User] = None,
    ) -> None:
        super().__init__(data, notify, fragment=fragment, user=user)


class AddBook(AddBookOptionalFragment):
    def __init__(
        self,
        data: MultiDict[str, Any],
        notify: Callable[[str, str], None],
        /,
        *,
        user: Optional[User] = None,
    ) -> None:
        super().__init__(data, notify, user=user)


class LookUpBook(HtmlOverHttpMixin, HttpPresenter, look_up_book.Presenter):
    def __init__(self, /, *, user: Optional[User] = None) -> None:
        self._ = l10n.translator_for(user.locale if user else "")
        self.meta = HtmlMeta()
        self.content = HtmlContent(
            "books/add#form", {"user": user, "_": self._, "data": {}, "errors": {}}
        )

    def not_an_isbn(self, isbn: str) -> None:
        self.content.set("isbn", isbn)
        self.content.set("errors", {"isbn": self._("not-an-isbn")})

    def book_not_found(self, isbn: Isbn) -> None:
        self.content.set("isbn", str(isbn))
        self.content.set("errors", {"isbn": self._("unknown-isbn")})

    def book(self, book: Book) -> None:
        self.content.set(
            "data",
            {
                "isbn": str(book.isbn),
                "title": str(book.title),
                "year": int(book.year),
                "authors": ", ".join([str(a) for a in book.authors]),
                "genres": ", ".join([str(g) for g in book.genres]),
                "cover": str(book.cover) if book.cover else "",
            },
        )


class DisplayBook(HtmlOverHttpMixin, HttpPresenter, display_book.Presenter):
    def __init__(
        self, notify: Callable[[str, str], None], /, *, user: Optional[User] = None
    ) -> None:
        self._ = l10n.translator_for(user.locale if user else "")
        self.meta = HtmlMeta(is_logged_in=bool(user))
        self.content = HtmlContent(
            "books/display", {"user": user, "_": self._, "book": {}}
        )
        self.__notify = notify

    def not_an_isbn(self, isbn: str) -> None:
        self.__notify(self._("not-an-isbn"), "warning")
        self.meta.see_other(url_for("books.search"))

    def book_not_found(self, isbn: Isbn) -> None:
        self.__notify(self._("book-does-not-exist"), "warning")
        self.meta.see_other(url_for("books.search"))

    def see_other(self, isbn: Isbn) -> None:
        self.meta.see_other(url_for("books.display", isbn=str(isbn)), permanent=True)

    def book(self, book: Book) -> None:
        self.content.set(
            "book",
            {
                "added_on": self._("date", date=book.added_on.to_date()),
                "updated_on": (
                    self._("date", date=book.updated_on.to_date())
                    if book.updated_on != book.added_on
                    else "-"
                ),
                "isbn": str(book.isbn),
                "title": str(book.title),
                "year": int(book.year),
                "authors": ", ".join([str(a) for a in book.authors]),
                "genres": ", ".join([str(g) for g in book.genres]),
                "cover": str(book.cover) if book.cover else "",
            },
        )


class UpdateBookForm(HtmlOverHttpMixin, HttpPresenter, display_book.Presenter):
    def __init__(
        self,
        isbn: str,
        notify: Callable[[str, str], None],
        /,
        *,
        user: Optional[User] = None,
    ) -> None:
        self._ = l10n.translator_for(user.locale if user else "")
        self.meta = HtmlMeta(is_logged_in=bool(user))
        self.content = HtmlContent(
            "books/update",
            {"user": user, "_": self._, "isbn": isbn, "data": {}, "errors": {}},
        )
        self.__notify = notify

    def not_an_isbn(self, isbn: str) -> None:
        self.__notify(self._("not-an-isbn"), "warning")
        self.meta.see_other(url_for("books.search"))

    def book_not_found(self, isbn: Isbn) -> None:
        self.__notify(self._("book-does-not-exist"), "warning")
        self.meta.see_other(url_for("books.search"))

    def see_other(self, isbn: Isbn) -> None:
        self.meta.see_other(url_for("books.display", isbn=str(isbn)), permanent=True)

    def book(self, book: Book) -> None:
        self.content.set(
            "data",
            {
                "isbn": str(book.isbn),
                "title": str(book.title),
                "year": int(book.year),
                "authors": ", ".join([str(a) for a in book.authors]),
                "genres": ", ".join([str(g) for g in book.genres]),
                "cover": str(book.cover) if book.cover else "",
            },
        )


class UpdateBookOptionalFragment(
    HtmlOverHttpMixin, HttpPresenter, update_book.Presenter
):
    def __init__(
        self,
        isbn: str,
        data: MultiDict[str, Any],
        notify: Callable[[str, str], None],
        /,
        *,
        fragment: str = "",
        user: Optional[User] = None,
    ) -> None:
        self._ = l10n.translator_for(user.locale if user else "")
        if fragment:
            self.meta = HtmxMeta(is_logged_in=bool(user))
        else:
            self.meta = HtmlMeta(is_logged_in=bool(user))
        self.content = HtmlContent(
            f"books/update#{fragment}",
            {"user": user, "_": self._, "isbn": isbn, "data": data, "errors": {}},
        )
        self.__notify = notify

    def bad_request(self, errors: update_book.Errors) -> None:
        self.content.set(
            "errors",
            {
                "title": translate_error(self._, errors.title),
                "year": translate_error(self._, errors.year),
                "authors": translate_error(self._, errors.authors),
            },
        )

    def book_not_found(self, isbn: Isbn) -> None:
        self.__notify(self._("book-not-found"), "warning")
        self.meta.see_other(url_for("books.search"))

    def book_updated(self, book: Book) -> None:
        self.__notify(self._("update-book-success"), "success")
        self.meta.see_other(url_for("books.display", isbn=str(book.isbn)))


class UpdateBookFragment(UpdateBookOptionalFragment):
    def __init__(
        self,
        fragment: str,
        isbn: str,
        data: MultiDict[str, Any],
        notify: Callable[[str, str], None],
        /,
        *,
        user: Optional[User] = None,
    ) -> None:
        super().__init__(isbn, data, notify, fragment=fragment, user=user)


class UpdateBook(UpdateBookOptionalFragment):
    def __init__(
        self,
        isbn: str,
        data: MultiDict[str, Any],
        notify: Callable[[str, str], None],
        /,
        *,
        user: Optional[User] = None,
    ) -> None:
        super().__init__(isbn, data, notify, user=user)


class ValidateBook(HtmlOverHttpMixin, HttpPresenter):
    def __init__(
        self, data: MultiDict[str, Any], /, *, user: Optional[User] = None
    ) -> None:
        if len(data.keys()) != 1:
            raise RuntimeError("Only one attribute can be validated at a time!")
        attribute = list(data.keys())[0]
        value = data[attribute]

        self._ = l10n.translator_for(user.locale if user else "")
        errors = self.__validate(attribute, value)
        context = {"user": user, "_": self._, "data": data, "errors": errors}

        self.meta = HtmlMeta()
        # We use `search` as it contains all fields as not required.
        self.content = HtmlContent(f"books/search#{attribute}", context)

    def __validate(self, attribute: str, value: str) -> dict[str, str]:
        errors: dict[str, str] = {}

        if not value:
            return errors

        if attribute == "isbn":
            errors["isbn"] = translate_error(
                self._, add_book.Errors(isbn=validators.isbn(value)).isbn
            )
        if attribute == "year":
            errors["year"] = translate_error(
                self._, add_book.Errors(year=validators.year(int(value))).year
            )

        return errors



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import Any, Optional
from urllib.parse import urlparse

from bl_hector.domain.administration.enumerations import Permissions
from bl_hector.domain.collection_management import errors
from bl_hector.interfaces.l10n import Translator


@dataclass
class User:
    id: str
    locale: str
    permissions: list[Permissions]

    def __getattribute__(self, name: str) -> Any:
        PREFIX = "can_"
        if name.startswith(PREFIX):
            return self.has_permission(name[len(PREFIX) :].upper())
        return super().__getattribute__(name)

    def has_permission(self, permission: str) -> bool:
        if not self.id:
            return False
        try:
            return Permissions[permission] in self.permissions
        except KeyError:
            return False


class Pager:
    # Should be divisible by 2 and 3 for rows to always be complete.
    PAGE_SIZE = 12

    def __init__(self, url: str) -> None:
        self.__url = urlparse(url)
        self.previous = self.__previous_page()
        self.next = self.__next_page()

    def __previous_page(self) -> str:
        page = self.__get_page()
        if page < 2:
            return ""
        return self.__set_page(page - 1)

    def __get_page(self) -> int:
        for q in self.__get_query():
            if q[0] == "page":
                if q[1].isnumeric():
                    return int(q[1])
        return 1

    def __get_query(self) -> list[list[str]]:
        return [q.split("=", 1) for q in self.__url.query.split("&")]

    def __set_page(self, page: int) -> str:
        if not page:
            return ""
        page_replaced = False
        query = []
        for q in self.__get_query():
            if q[0] == "page":
                query.append(f"page={page}")
                page_replaced = True
            else:
                query.append("=".join(q))
        if not page_replaced:
            query.append(f"page={page}")
        return self.__url._replace(query="&".join(query)).geturl()

    def __next_page(self) -> str:
        return self.__set_page(self.__get_page() + 1)


def translate_error(
    translator: Translator, error: Optional[errors.InvalidValue] = None
) -> str:
    if not error:
        return ""

    # Dispatching errors based on class is not very LSP!?
    if isinstance(error, errors.StringTooShort):
        return translator("string-too-short", min=error.MIN)
    if isinstance(error, errors.StringTooLong):
        return translator("string-too-long", max=error.MAX)
    if isinstance(error, errors.NumberTooSmall):
        return translator("number-too-small", min=error.MIN)
    if isinstance(error, errors.NumberTooBig):
        return translator("number-too-big", max=error.MAX)
    if isinstance(error, errors.UnknownBook):
        return translator("unknown-book")
    if isinstance(error, errors.NotAnIsbn):
        return translator("not-an-isbn")
    if isinstance(error, errors.MissingAuthor):
        return translator("missing-author")
    if isinstance(error, errors.BeforeCreationOfIsbn):
        return translator("before-creation-of-isbn", year=str(error.year))
    if isinstance(error, errors.InvalidValue):
        return translator("incorrect-value")

    return translator("unknown-error")
